/*
 * fft2pc.c -- Character set translator for Final Fantasy Tactics by
 *	Squaresoft.  This is not intended for copyright infringement.
 *
 * Copyright 2000, George Greer <greerga@fftactics.org>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include "mmaplib.h"

static const char *program = __FILE__ ": Final Fantasy Tactics character set translator.";
static const char *version = "v0.5 (Build: "__DATE__" "__TIME__")";
static const char *copyright = "Copyright 2000, George Greer <greerga@fftactics.org>";
static const char *license = "Licensed under GNU General Public License (GPL) v2 or later.";

#define FALSE	(0)
#define TRUE	(!FALSE)

struct options_s {
  unsigned int rawmultibyte : 1;
  unsigned int printhtml : 1;
};

/* ------------------------------------------------------------------------------------------------------------------------------ */

/*
 * I know this could definitely be more efficient than a linear scan.
 *
 * Additional Hiragana/Katakana table listing from Gelwan (Paul Le) on www.ffnet.net boards.
 */
struct {
  unsigned char match;
  struct {
    unsigned char multibyte;
    const char *text;
    const char *html;
  } submatch[256];			/* Yeah, well, :P */
  const char *text;
  const char *html;
} xlate[] = {
	/* Hiragana */
  { 0x3E, {},	"!",			NULL			},
  { 0x3F, {},	"a",			"&#12354;"		},
  { 0x40, {},	"?",			NULL			},
  { 0x41, {},	"i",			"&#12356;"		},
  { 0x42, {},	"+",			NULL			},
  { 0x43, {},	"u",			"&#12358;"		},
  { 0x44, {},	"/",			NULL			},
  { 0x45, {},	"e",			"&#12360;"		},
  { 0x46, {},	":",			NULL			},
  { 0x47, {},	"o",			"&#12362;"		},
  { 0x48, {},	"ka",			"&#12363;"		},
  { 0x49, {},	"ga",			"&#12364;"		},
  { 0x4A, {},	"ki",			"&#12365;"		},
  { 0x4B, {},	"gi",			"&#12366;"		},
  { 0x4C, {},	"ku",			"&#12367;"		},
  { 0x4D, {},	"gu",			"&#12368;"		},
  { 0x4E, {},	"ke",			"&#12369;"		},
  { 0x4F, {},	"ge",			"&#12370;"		},
  { 0x50, {},	"ko",			"&#12371;"		},
  { 0x51, {},	"go",			"&#12372;"		},
  { 0x52, {},	"sa",			"&#12373;"		},
  { 0x53, {},	"za",			"&#12374;"		},
  { 0x54, {},	"shi",			"&#12375;"		},
  { 0x55, {},	"ji",			"&#12376;"		},
  { 0x56, {},	"su",			"&#12377;"		},
  { 0x57, {},	"zu",			"&#12378;"		},
  { 0x58, {},	"se",			"&#12379;"		},
  { 0x59, {},	"ze",			"&#12380;"		},
  { 0x5A, {},	"so",			"&#12381;"		},
  { 0x5B, {},	"zo"			"&#12382;"		},
  { 0x5C, {},	"ta"			"&#12383;"		},
  { 0x5D, {},	"da"			"&#12384;"		},
  { 0x5E, {},	"chi"			"&#12385;"		},
  { 0x5F, {},	".",			NULL			},
  { 0x60, {},	"xtsu"			"&#12387;"		},
  { 0x61, {},	"tsu"			"&#12388;"		},
  { 0x62, {},	"tzu"			"&#12389;"		},
  { 0x63, {},	"te"			"&#12390;"		},
  { 0x64, {},	"de"			"&#12391;"		},
  { 0x65, {},	"to"			"&#12392;"		},
  { 0x66, {},	"do"			"&#12393;"		},
  { 0x67, {},	"na"			"&#12394;"		},
  { 0x68, {},	"ni"			"&#12395;"		},
  { 0x69, {},	"nu",			"&#12396;"		},
  { 0x6A, {},	"ne",			"&#12397;"		},
  { 0x6B, {},	"no",			"&#12398;"		},
  { 0x6C, {},	"ha",			"&#12399;"		},
  { 0x6D, {},	"ba",			"&#12400;"		},
  { 0x6E, {},	"pa",			"&#12401;"		},
  { 0x6F, {},	"hi",			"&#12402;"		},
  { 0x70, {},	"bi",			"&#12403;"		},
  { 0x71, {},	"pi",			"&#12404;"		},
  { 0x72, {},	"fu",			"&#12405;"		},
  { 0x73, {},	"bu",			"&#12406;"		},
  { 0x74, {},	"pu",			"&#12407;"		},
  { 0x75, {},	"he",			"&#12408;"		},
  { 0x76, {},	"be",			"&#12409;"		},
  { 0x77, {},	"pe",			"&#12410;"		},
  { 0x78, {},	"ho",			"&#12411;"		},
  { 0x79, {},	"bo",			"&#12412;"		},
  { 0x7A, {},	"po",			"&#12413;"		},
  { 0x7B, {},	"ma",			"&#12414;"		},
  { 0x7C, {},	"mi",			"&#12415;"		},
  { 0x7D, {},	"mu",			"&#12416;"		},
  { 0x7E, {},	"me",			"&#12417;"		},
  { 0x7F, {},	"mo",			"&#12418;"		},
  { 0x80, {},	"a",			"&#12419;"		},
  { 0x81, {},	"ya",			"&#12420;"		},
  { 0x82, {},	"u",			"&#12421;"		},
  { 0x83, {},	"yu",			"&#12422;"		},
  { 0x84, {},	"o",			"&#12423;"		},
  { 0x85, {},	"yo",			"&#12424;"		},
  { 0x86, {},	"ra",			"&#12425;"		},
  { 0x87, {},	"ri",			"&#12426;"		},
  { 0x88, {},	"ru",			"&#12427;"		},
  { 0x89, {},	"re",			"&#12428;"		},
  { 0x8A, {},	"ro",			"&#12429;"		},
  { 0x8B, {},	"",			"&#12539;"		},
  { 0x8C, {},	"wa",			"&#12431;"		},
  { 0x8D, {},	"(",			"&#65288;"		},
  { 0x8E, {},	")",			"&#65289;"		},
  { 0x8F, {},	"o",			"&#12434;"		},
  { 0x90, {},	"n",			"&#12435;"		},
	/* Katakana */
  { 0x91, {},	"\"",			NULL			},
  { 0x92, {},	"a",			"&#12450;"		},
  { 0x93, {},	"'",			NULL			},
  { 0x94, {},	"i",			"&#12452;"		},
  { 0x95, {},	"u",			"&#12453;"		},
  { 0x96, {},	"u",			"&#12454;"		},
  { 0x97, {},	"e",			"&#12455;"		},
  { 0x98, {},	"e",			"&#12456;"		},
  { 0x99, {},	"o",			"&#12457;"		},
  { 0x9A, {},	"o",			"&#12458;"		},
  { 0x9B, {},	"ka",			"&#12459;"		},
  { 0x9C, {},	"ga",			"&#12460;"		},
  { 0x9D, {},	"ki",			"&#12461;"		},
  { 0x9E, {},	"gi",			"&#12462;"		},
  { 0x9F, {},	"ku",			"&#12463;"		},
  { 0xA0, {},	"gu",			"&#12464;"		},
  { 0xA1, {},	"ke",			"&#12465;"		},
  { 0xA2, {},	"ge",			"&#12466;"		},
  { 0xA3, {},	"ko",			"&#12467;"		},
  { 0xA4, {},	"go",			"&#12468;"		},
  { 0xA5, {},	"sa",			"&#12469;"		},
  { 0xA6, {},	"za",			"&#12470;"		},
  { 0xA7, {},	"shi",			"&#12471;"		},
  { 0xA8, {},	"ji",			"&#12472;"		},
  { 0xA9, {},	"su",			"&#12473;"		},
  { 0xAA, {},	"zu",			"&#12474;"		},
  { 0xAB, {},	"se",			"&#12475;"		},
  { 0xAC, {},	"ze",			"&#12476;"		},
  { 0xAD, {},	"so",			"&#12477;"		},
  { 0xAE, {},	"zo",			"&#12478;"		},
  { 0xAF, {},	"ta",			"&#12479;"		},
  { 0xB0, {},	"da",			"&#12480;"		},
  { 0xB1, {},	"chi",			"&#12481;"		},
  { 0xB2, {},	"[note]",		NULL			},
  { 0xB3, {},	"xtsu",			"&#12483;"		},
  { 0xB4, {},	"tsu",			"&#12484;"		},
  { 0xB5, {},	"*",			NULL			},
  { 0xB6, {},	"te",			"&#12486;"		},
  { 0xB7, {},	"de",			"&#12487;"		},
  { 0xB8, {},	"to",			"&#12488;"		},
  { 0xB9, {},	"do",			"&#12489;"		},
  { 0xBA, {},	"na",			"&#12490;"		},
  { 0xBB, {},	"ni",			"&#12491;"		},
  { 0xBC, {},	"nu",			"&#12492;"		},
  { 0xBD, {},	"ne",			"&#12493;"		},
  { 0xBE, {},	"no",			"&#12494;"		},
  { 0xBF, {},	"ha",			"&#12495;"		},
  { 0xC0, {},	"ba",			"&#12496;"		},
  { 0xC1, {},	"pa",			"&#12497;"		},
  { 0xC2, {},	"hi",			"&#12498;"		},
  { 0xC3, {},	"bi",			"&#12499;"		},
  { 0xC4, {},	"pi",			"&#12500;"		},
  { 0xC5, {},	"fu",			"&#12501;"		},
  { 0xC6, {},	"bu",			"&#12502;"		},
  { 0xC7, {},	"pu",			"&#12503;"		},
  { 0xC8, {},	"he",			"&#12504;"		},
  { 0xC9, {},	"be",			"&#12505;"		},
  { 0xCA, {},	"pe",			"&#12506;"		},
  { 0xCB, {},	"ho",			"&#12507;"		},
  { 0xCC, {},	"bo",			"&#12508;"		},
  { 0xCD, {},	"po",			"&#12509;"		},
  { 0xCE, {},	"ma",			"&#12510;"		},
  { 0xCF, {},	"mi",			"&#12511;"		},
  { 0xD1, {
    { 0x00,	"mu",			"&#12512;"		},
    { 0x01,	"me",			"&#12513;"		},
    { 0x02,	"mo",			"&#12514;"		},
    { 0x03,	"a",			"&#12515;"		},
    { 0x04,	"a",			"&#12516;"		},
    { 0x05,	"u",			"&#12517;"		},
    { 0x06,	"u",			"&#12518;"		},
    { 0x07,	"o",			"&#12519;"		},
    { 0x08,	"o",			"&#12520;"		},
    { 0x09,	"ra",			"&#12521;"		},
    { 0x0A,	"ri",			"&#12522;"		},
    { 0x0B,	"ru",			"&#12523;"		},
    { 0x0C,	"re",			"&#12524;"		},
    { 0x0D,	"ro",			"&#12525;"		},
    { 0x0E,	"xwa",			"&#12526;"		},
    { 0x0F,	"wa",			"&#12527;"		},
    { 0x10,	"[recursive arrow]",	NULL			},
    { 0x11,	"*",			"&#8251;"		},
    { 0x12,	"o",			"&#12530;"		},
    { 0x13,	"n",			"&#12531;"		},
    { 0x14,	"v",			"&#12532;"		},
    { 0x15,	"xka",			"&#12533;"		},
    { 0x16,	"xke",			"&#12534;"		},
    { 0x17,	"--",			"&#8213;"		},
    { 0x18,	"{",			"&#12300;"		},
    { 0x19,	",",			"&#12289;"		},
    { 0x1A,	"!",			"&#65281;"		},
    { 0x1B,	"...",			"..."			},
    { 0x1C,	".",			"&#12290;"		},
    { 0x1D,	"-"					},
    { 0x1E,	"+"					},
    { 0x1F,	""					},
    { 0x20,	""					},
    { 0x21,	"[half circle]"				},
    { 0x22,	"[half circle]"				},
    { 0x23,	"="					},
    { 0x24,	"!="					},
    { 0x25,	">"					},
    { 0x26,	"<"					},
    { 0x27,	">="					},
    { 0x28,	"<="					},
    { 0x29,	"*"					},
    { 0x2A,	"*"					},
    { 0x2B,	"*"					},
    { 0x2C,	"*"					},
    { 0x2D,	"*"					},
    { 0x2E,	"*"					},
    { 0x2F,	"*"					},
    { 0x30,	"*"					},
    { 0x31,	"*"					},
    { 0x32,	"*"					},
    { 0x33,	NULL,			"&#21091;"	},
    { 0x34,	NULL,			"&#19968;"	},
    { 0x35,	NULL,			"&#20057;"	},
    { 0x36,	NULL,			"&#19971;"	},
    { 0x37,	NULL,			"&#19969;"	},
    { 0x38,	NULL,			"&#20061;"	},
    { 0x39,	NULL,			"&#20102;"	},
    { 0x3A,	NULL,			"&#24974;"	},
    { 0x3B,	NULL,			"&#20154;"	},
    { 0x3C,	NULL,			"&#20837;"	},
    { 0x3D,	NULL,			"&#20843;"	},
    { 0x3E,	NULL,			"&#20992;"	},
    { 0x3F,	NULL,			"&#21147;"	},
    { 0x40,	NULL,			"&#21313;"	},
    { 0x41,	NULL,			"&#19979;"	},
    { 0x42,	NULL,			"&#19977;"	},
    { 0x43,	NULL,			"&#19978;"	},
    { 0x44,	NULL,			"&#19976;"	},
    { 0x45,	NULL,			"&#19975;"	},
    { 0x46,	NULL,			"&#19982;"	},
    { 0x47,	NULL,			"&#20037;"	},
    { 0x48,	NULL,			"&#20024;"	},
    { 0x49,	NULL,			"&#20062;"	},
    { 0x4A,	NULL,			"&#20063;"	},
    { 0x4B,	NULL,			"&#20129;"	},
    { 0x4C,	NULL,			"&#20961;"	},
    { 0x4D,	NULL,			"&#20995;"	},
    { 0x4E,	NULL,			"&#21315;"	},
    { 0x4F,	NULL,			"&#39151;"	},
    { 0x50,	NULL,			"&#22303;"	},
    { 0x51,	NULL,			"&#22763;"	},
    { 0x52,	NULL,			"&#22805;"	},
    { 0x53,	NULL,			"&#22823;"	},
    { 0x54,	NULL,			"&#22899;"	},
    { 0x55,	NULL,			"&#23376;"	},
    { 0x56,	NULL,			"&#23544;"	},
    { 0x57,	NULL,			"&#23567;"	},
    { 0x58,	NULL,			"&#23665;"	},
    { 0x59,	NULL,			"&#24029;"	},
    { 0x5A,	NULL,			"&#24037;"	},
    { 0x5B,	NULL,			"&#24050;"	},
    { 0x5C,	NULL,			"&#24178;"	},
    { 0x5D,	NULL,			"&#24339;"	},
    { 0x5E,	NULL,			"&#12293;"	},
    { 0x5F,	NULL,			"&#27833;"	},
    { 0x60,	NULL,			"&#31085;"	},
    { 0x61,	NULL,			"&#22855;"	},
    { 0x62,	NULL,			"&#36321;"	},
    { 0x63,	NULL,			"&#28436;"	},
    { 0x64,	NULL,			"&#19981;"	},
    { 0x65,	NULL,			"&#20013;"	},
    { 0x66,	NULL,			"&#20104;"	},
    { 0x67,	NULL,			"&#20116;"	},
    { 0x68,	NULL,			"&#20114;"	},
    { 0x69,	NULL,			"&#20117;"	},
    { 0x6A,	NULL,			"&#20171;"	},
    { 0x6B,	NULL,			"&#20167;"	},
    { 0x6C,	NULL,			"&#20170;"	},
    { 0x6D,	NULL,			"&#20161;"	},
    { 0x6E,	NULL,			"&#20869;"	},
    { 0x6F,	NULL,			"&#20803;"	},
    { 0x70,	NULL,			"&#20844;"	},
    { 0x71,	NULL,			"&#20845;"	},
    { 0x72,	NULL,			"&#20870;"	},
    { 0x73,	NULL,			"&#20887;"	},
    { 0x74,	NULL,			"&#20982;"	},
    { 0x75,	NULL,			"&#20999;"	},
    { 0x76,	NULL,			"&#20998;"	},
    { 0x77,	NULL,			"&#21250;"	},
    { 0x78,	NULL,			"&#21270;"	},
    { 0x79,	NULL,			"&#24040;"	},
    { 0x7A,	NULL,			"&#21305;"	},
    { 0x7B,	NULL,			"&#21320;"	},
    { 0x7C,	NULL,			"&#21380;"	},
    { 0x7D,	NULL,			"&#21452;"	},
    { 0x7E,	NULL,			"&#21453;"	},
    { 0x7F,	NULL,			"&#21451;"	},
    { 0x80,	NULL,			"&#22826;"	},
    { 0x81,	NULL,			"&#22825;"	},
    { 0x82,	NULL,			"&#23569;"	},
    { 0x83,	NULL,			"&#24187;"	},
    { 0x84,	NULL,			"&#24341;"	},
    { 0x85,	NULL,			"&#24515;"	},
    { 0x86,	NULL,			"&#25144;"	},
    { 0x87,	NULL,			"&#25163;"	},
    { 0x88,	NULL,			"&#25903;"	},
    { 0x89,	NULL,			"&#25991;"	},
    { 0x8A,	NULL,			"&#26007;"	},
    { 0x8B,	NULL,			"&#26041;"	},
    { 0x8C,	NULL,			"&#26085;"	},
    { 0x8D,	NULL,			"&#26376;"	},
    { 0x8E,	NULL,			"&#26408;"	},
    { 0x8F,	NULL,			"&#27424;"	},
    { 0x90,	NULL,			"&#27490;"	},
    { 0x91,	NULL,			"&#27604;"	},
    { 0x92,	NULL,			"&#27611;"	},
    { 0x93,	NULL,			"&#27700;"	},
    { 0x94,	NULL,			"&#28779;"	},
    { 0x95,	NULL,			"&#29226;"	},
    { 0x96,	NULL,			"&#29238;"	},
    { 0x97,	NULL,			"&#29255;"	},
    { 0x98,	NULL,			"&#29275;"	},
    { 0x99,	NULL,			"&#29356;"	},
    { 0x9A,	NULL,			"&#29579;"	},
    { 0x9B,	NULL,			"&#20581;"	},
    { 0x9C,	NULL,			"&#24247;"	},
    { 0x9D,	NULL,			"&#32908;"	},
    { 0x9E,	NULL,			"&#29359;"	},
    { 0x9F,	NULL,			"&#23629;"	},
    { 0xA0,	NULL,			"&#25943;"	},
    { 0xA1,	NULL,			"&#25105;"	},
    { 0xA2,	NULL,			"&#30331;"	},
    { 0xA3,	NULL,			"&#37682;"	},
    { 0xA4,	NULL,			"&#19992;"	},
    { 0xA5,	NULL,			"&#19990;"	},
    { 0xA6,	NULL,			"&#20027;"	},
    { 0xA7,	NULL,			"&#20197;"	},
    { 0xA8,	NULL,			"&#20181;"	},
    { 0xA9,	NULL,			"&#20185;"	},
    { 0xAA,	NULL,			"&#20182;"	},
    { 0xAB,	NULL,			"&#20195;"	},
    { 0xAC,	NULL,			"&#20184;"	},
    { 0xAD,	NULL,			"&#20196;"	},
    { 0xAE,	NULL,			"&#20804;"	},
    { 0xAF,	NULL,			"&#20889;"	},
    { 0xB0,	NULL,			"&#20966;"	},
    { 0xB1,	NULL,			"&#20986;"	},
    { 0xB2,	NULL,			"&#21152;"	},
    { 0xB3,	NULL,			"&#21253;"	},
    { 0xB4,	NULL,			"&#21271;"	},
    { 0xB5,	NULL,			"&#21322;"	},
    { 0xB6,	NULL,			"&#21344;"	},
    { 0xB7,	NULL,			"&#21435;"	},
    { 0xB8,	NULL,			"&#21454;"	},
    { 0xB9,	NULL,			"&#21487;"	},
    { 0xBA,	NULL,			"&#21151;"	},
    { 0xBB,	NULL,			"&#21477;"	},
    { 0xBC,	NULL,			"&#21476;"	},
    { 0xBD,	NULL,			"&#21495;"	},
    { 0xBE,	NULL,			"&#21490;"	},
    { 0xBF,	NULL,			"&#21496;"	},
    { 0xC0,	NULL,			"&#21484;"	},
    { 0xC1,	NULL,			"&#21488;"	},
    { 0xC2,	NULL,			"&#21491;"	},
    { 0xC3,	NULL,			"&#22235;"	},
    { 0xC4,	NULL,			"&#22234;"	},
    { 0xC5,	NULL,			"&#22311;"	},
    { 0xC6,	NULL,			"&#20908;"	},
    { 0xC7,	NULL,			"&#22806;"	},
    { 0xC8,	NULL,			"&#22830;"	},
    { 0xC9,	NULL,			"&#22833;"	},
    { 0xCA,	NULL,			"&#22900;"	},
    { 0xCB,	NULL,			"&#23611;"	},
    { 0xCC,	NULL,			"&#24038;"	},
    { 0xCD,	NULL,			"&#24066;"	},
    { 0xCE,	NULL,			"&#24067;"	},
    { 0xCF,	NULL,			"&#24179;"	},
    { 0xD0,	NULL,			"&#24188;"	},
    { 0xD1,	NULL,			"&#24195;"	},
    { 0xD2,	NULL,			"&#24517;"	},
    { 0xD3,	NULL,			"&#25171;"	},
    { 0xD4,	NULL,			"&#25173;"	},
    { 0xD5,	NULL,			"&#26087;"	},
    { 0xD6,	NULL,			"&#26413;"	},
    { 0xD7,	NULL,			"&#26412;"	},
    { 0xD8,	NULL,			"&#26411;"	},
    { 0xD9,	NULL,			"&#26410;"	},
    { 0xDA,	NULL,			"&#27491;"	},
    { 0xDB,	NULL,			"&#27665;"	},
    { 0xDC,	NULL,			"&#27704;"	},
    { 0xDD,	NULL,			"&#27703;"	},
    { 0xDE,	NULL,			"&#29273;"	},
    { 0xDF,	NULL,			"&#29577;"	},
    { 0xE0,	NULL,			"&#29976;"	},
    { 0xE1,	NULL,			"&#29983;"	},
    { 0xE2,	NULL,			"&#29992;"	},
    { 0xE3,	NULL,			"&#30000;"	},
    { 0xE4,	NULL,			"&#30002;"	},
    { 0xE5,	NULL,			"&#30003;"	},
    { 0xE6,	NULL,			"&#30001;"	},
    { 0xE7,	NULL,			"&#30333;"	},
    { 0xE8,	NULL,			"&#30382;"	},
    { 0xE9,	NULL,			"&#30446;"	},
    { 0xEA,	NULL,			"&#30690;"	},
    { 0xEB,	NULL,			"&#30707;"	},
    { 0xEC,	NULL,			"&#31034;"	},
    { 0xED,	NULL,			"&#31036;"	},
    { 0xEE,	NULL,			"&#31348;"	},
    { 0xEF,	NULL,			"&#31435;"	},
#if 0
    { 0xF0	},
    { 0xF1	},
    { 0xF2	},
    { 0xF3	},
    { 0xF4	},
    { 0xF5	},
    { 0xF6	},
    { 0xF7	},
    { 0xF8	},
    { 0xF9	},
    { 0xFA	},
    { 0xFB	},
    { 0xFC	},
    { 0xFD	},
    { 0xFE	},
    { 0xFF	},
#endif
  }, NULL },
  { 0xD2, {},	NULL					},
  { 0xD9, {
    { 0xB5,	"[infinity]"				},
    { 0xB6,	"."					},
    { 0xB7,	"&"					},
    { 0xB8,	"%"					},
    { 0xB9,	"[O]"					},	/* The 'O' controller button. */
    { 0xBA,	"<="					},	/* Left arrow on name input screen. */
    { 0xBB,	"=>"					},	/* Right arrow on name input screen. */
    { 0xBC,	""					},
    { 0xBD,	":"					},
    { 0xBE,	"("					},
    { 0xBF,	")"					},
    { 0xC0,	"\""					},
    { 0xC1,	"'"					},
    { 0xC2,	"["					},	/* Version 1 bracket */
    { 0xC3,	"]"					},	/* Version 1 bracket */
    { 0xC4,	"}"					},	/* Version 2 bracket */
    { 0xC5,	"~"					},
    { 0xC6,	"/"					},
    { 0xC7,	"[T]"					},	/* The 'Triangle' controller button. */
    { 0xC8,	"[S]"					},	/* The 'Square' controller button. */
    { 0xC9,	"?"					},
    { 0xCA,	"[heart]"				},
    { 0xCB,	"I"					},
    { 0xCC,	"II"					},
    { 0xCD,	"III"					},
    { 0xCE,	"IV"					},
    { 0xCF,	"V"					},
    { 0xD0,	"Aries"					},
    { 0xD1,	"Taurus"				},
    { 0xD2,	"Gemini"				},
    { 0xD3,	"Cancer"				},
    { 0xD4,	"Leo"					},
    { 0xD5,	"Virgo"					},
    { 0xD6,	"Libra"					},
    { 0xD7,	"Scorpio"				},
    { 0xD8,	"Sagittarius"				},
    { 0xD9,	"Capricorn"				},
    { 0xDA,	"Aquarius"				},
    { 0xDB,	"Pisces"				},
    { 0xDC,	"Sagittarius"				},
  }, NULL },
  { 0xDA, {
    { 0x00,	"Aries"					},
    { 0x01,	"Taurus"				},
    { 0x02,	"Gemini"				},
    { 0x03,	"Cancer"				},
    { 0x04,	"Leo"					},
    { 0x05,	"Virgo"					},
    { 0x06,	"Libra"					},
    { 0x07,	"Scorpio"				},
    { 0x08,	"Sagittarius"				},
    { 0x09,	"Capricorn"				},
    { 0x0A,	"Aquarius"				},
    { 0x0B,	"Pisces"				},
    { 0x0C,	"Serpentarius"				},
    { 0x70,	"="					},
    { 0x71,	"$"					},
    { 0x72,	""					},
    { 0x73,	" "					},
    { 0x74,	","					},
    { 0x75,	";"					},
    { 0x76,	"'"					},
    { 0x77,	"\""					},
  }, NULL },
  { 0xE0, {},	"[Ramza]"				},	/* Obviously this "variable" (0xE0-0xEB) stuff seems overloaded...hm. */
#if 0
  { 0xE1, {},	NULL					},
#endif
  { 0xE2, {
#if 0
    { 0x05,	"[Thoughts]"				},	/* The non-boxed message texts like before Magic City Gariland battle. */
    { 0x0F,	"[/Thoughts]"				},
#endif
  }, NULL },
  { 0xE3, {
    { 0x00,	"[Normal Text]"				},
    { 0x04,	"[Gray Text]"				},
    { 0x08,	"[Red Text]"				},
  }, NULL },
#if 0
  { 0xE4, {},	NULL					},	/* Seems to be 1-byte. */
#endif
  { 0xE5, {},	NULL					},	/* Seems to be 2-byte. */
  { 0xE6, {},	NULL					},	/* Seems to be 2-byte. */
  { 0xE7, {
    { 0x02,	"  "					},	/* Name input screen uses for spacing. */
    { 0x03,	"  "					},	/* Name input screen uses for spacing. */
    { 0x04,	"  "					},	/* Name input screen uses for spacing. */
    { 0x16,	"  "					},	/* Name input screen uses for spacing. */
    { 0x17,	"  "					},	/* Name input screen uses for spacing. */
    { 0x2A,	"  "					},	/* Name input screen uses for spacing. */
    { 0x2B,	"  "					},	/* Name input screen uses for spacing. */
    { 0x3E,	"  "					},	/* Name input screen uses for spacing. */
    { 0x3F,	"  "					},	/* Name input screen uses for spacing. */
    { 0x52,	"  "					},	/* Name input screen uses for spacing. */
    { 0x53,	"  "					},	/* Name input screen uses for spacing. */
    { 0x66,	"  "					},	/* Name input screen uses for spacing. */
    { 0x67,	"  "					},	/* Name input screen uses for spacing. */
    { 0x68,	"  "					},	/* Name input screen uses for spacing. */
    { 0x78,	"  "					},	/* Name input screen uses for spacing. */
    { 0x79,	"  "					},	/* Name input screen uses for spacing. */
    { 0x7A,	"  "					},	/* Name input screen uses for spacing. */
    { 0x7B,	"  "					},	/* Name input screen uses for spacing. */
    { 0x8E,	"  "					},	/* Name input screen uses for spacing. */
    { 0x8F,	"  "					},	/* Name input screen uses for spacing. */
    { 0xA2,	"  "					},	/* Name input screen uses for spacing. */
    { 0xA3,	"  "					},	/* Name input screen uses for spacing. */
    { 0xA4,	"  "					},	/* Name input screen uses for spacing. */
    { 0xB6,	"  "					},	/* Name input screen uses for spacing. */
    { 0xB7,	"  "					},	/* Name input screen uses for spacing. */
    { 0xC9,	"  "					},	/* Name input screen uses for spacing. */
    { 0xCA,	"  "					},	/* Name input screen uses for spacing. */
  }, NULL	},
  { 0xE8, {},	NULL					},	/* Seems to be 2-byte. */
#if 0
  { 0xE9, {},	NULL					},	/* Seems to be 1-byte. */
  { 0xEA, {},	NULL					},	/* Seems to be 1-byte. */
  { 0xEB, {},	NULL					},	/* Seems to be 1-byte. */
#endif
  { 0xEC, {},	"[<special>]"				},	/* Battle character Brave Story profile picture number. */
  { 0xF0, {},	NULL					},	/* Compression keys. */
  { 0xF5, {},	NULL					},	/* Used for Brave Story pictures. */
  { 0xF6, {},	NULL					},	/* Used for Brave Story pictures. */
  { 0xF8, {},	"\n",			"<BR>\n"	},	/* In-text */
  { 0xFA, {},	" "					},
//{ 0xFD, {},	"???"					},	/* "Carve <FD> Model" */
  { 0xFE, {},	"\n\n",			"\n\n<P>"	},	/* End-of-text */
  { 0x00, {},	NULL					},	/* Guard value. */
};

int fft2pc(const unsigned char *input, int length, struct options_s *options)
{
  int rpos;

  if (options->printhtml)
    fputs(
	"<HTML>\n"
	"<HEAD>\n"
	"  <META NAME=\"Content-type\" VALUE=\"Shift_JIS\">\n"
	"</HEAD>\n"
	"<BODY>\n", stdout);

  for (rpos = 0; rpos < length; rpos++) {
    if (input[rpos] <= 0x9)
      putchar('0' + input[rpos]);
    else if (input[rpos] <= 0x23)
      putchar('A' + (input[rpos] - 0x0A));
    else if (input[rpos] <= 0x3D)
      putchar('a' + (input[rpos] - 0x24));
    else if (input[rpos] == 0xEC)
      printf("[Picture #%u]", input[++rpos]);
    else {
      int first, second;

      for (first = 0; xlate[first].text || xlate[first].match; first++) {
        if (input[rpos] != xlate[first].match)
          continue;

        if (options->rawmultibyte) {
          if (xlate[first].text)
            printf("<%0X>", input[rpos]);
          else {
            rpos++;
            printf("<%0X/%0X>", input[rpos - 1], input[rpos]);
          }
          break;
        }

        if (!xlate[first].submatch[0].text && !xlate[first].text && !options->printhtml) {
          rpos++;
          printf("<%0X/%0X>", input[rpos - 1], input[rpos]);
          break;
        }

        if (xlate[first].submatch[0].text) {
          rpos++;
          for (second = 0; xlate[first].submatch[second].text; second++) {
            if (input[rpos] != xlate[first].submatch[second].multibyte)
              continue;

            fputs(options->printhtml && xlate[first].submatch[second].html ?
		xlate[first].submatch[second].html : xlate[first].submatch[second].text, stdout);

            break;
          }
          if (!xlate[first].submatch[second].text && !xlate[first].text && !options->printhtml)
            printf("<%0X/%0X>", input[rpos - 1], input[rpos]);
          else
            break;
        }

        if (xlate[first].text)
          fputs(options->printhtml && xlate[first].html ? xlate[first].html : xlate[first].text, stdout);

        break;
      }
      if (!xlate[first].text && !xlate[first].match && !options->printhtml)
        printf("<%0X>", input[rpos]);
    } /* if */
  } /* for */

  if (options->printhtml)
    fputs("\n</BODY></HTML>\n", stdout);

  return rpos;
}

/* ------------------------------------------------------------------------------------------------------------------------------ */

struct option /* {
  const char *name;
  int has_arg;
  int *flag;
  int val;
} */ long_options[] = {
  { "help",     FALSE, NULL, 0 },
  { "version",  FALSE, NULL, 1 },
  { 0, 0, 0, 0 }
};

int print_help(char *progname)
{
  printf("%s [-r] [-h] filename\n"
	 "Use -V or --version for version and copyright information.\n"
	 "%s --help	This screen.\n",
	progname, progname);
  return EXIT_FAILURE;
}

int print_version(char *progname)
{
  printf("%s\n\t%s\n\t%s\n\n%s: version %s\n", program, copyright, license, progname, version);
  return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
  void *handle;
  char *filename;
  int gto;
  struct options_s options = {
    rawmultibyte:	FALSE,
  };

  while ((gto = getopt_long(argc, argv, "hrV", long_options, NULL)) >= 0) {
    switch (gto) {
    case 0:
      return print_help(argv[0]);
    case 1:
    case 'V':
      return print_version(argv[0]);
    case 'h':
      options.printhtml = TRUE;
      break;
    case 'r':
      options.rawmultibyte = TRUE;
      break;

    case ':':
      fprintf(stderr, "Option -%c requires an extra argument.\n", optopt);
      return EXIT_FAILURE;
    case '?':
      fprintf(stderr, "Unknown option -%c.\n", optopt);
      return EXIT_FAILURE;
    default:
      fputs("Unknown error condition.\n", stderr);
      return EXIT_FAILURE;
    }
  }

  if (options.rawmultibyte && options.printhtml) {
    fputs("Cannot do both raw multi-byte and HTML printing.\n", stderr);
    exit(1);
  }

  if (argc == optind)
    fputs("Program requires a filename to translate.\nTry '--help' for more information.\n", stderr);
  else {
    while (optind < argc) {
      filename = argv[optind];
      if (!(handle = mapopen(filename)))
        error(filename);
      fft2pc(mapdata(handle), maplength(handle), &options);
      mapclose(handle);
      optind++;
    }
    return EXIT_SUCCESS;
  }
  return EXIT_FAILURE;
}
